(* $Id: godi_update.ml 711 2007-07-30 22:37:57Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Godi_types
open Godi_exec
open Godi_query
open Godi_versions
open Printf

(* 1. available.new ist eine Temporrdatei -> /tmp/xxxx/available.new
 *    Wenn Download schiefgeht: nicht schlimm, Fehlermeldung ausgeben,
 *    nichts in available_current eintragen.
 * 2. Variable available_current:
 *    - enthlt auch den Prfix des Servers
 *    - Wenn Paket mehrfach da: hhere Versionen schlagen niedrigere.
 *      Sonst nach Reihenfolge vorgehen.
 *)

let available_current = Hashtbl.create 200;;


let words_re = Str.regexp "[ \t\r\n]+";;

let parse_available con site file =
  let out = printmsg con in
  let f = 
    try
      open_in file
    with
	Sys_error msg ->
	  user_error("System error: " ^ msg)
  in
  let l = ref [] in
  let n = ref 0 in
  ( try
      while true do
	incr n;
	let line = input_line f in
	let words = Str.split words_re line in
	match words with
	    [ name; dir; version ] ->
	      l := (name, (dir, version, site)) :: !l
	  | _ ->
	      out (sprintf "Warning: Parsing error in line %d of %s (site %s)\n"
		     !n file site);
      done;
      assert false
    with
	End_of_file ->
	  close_in f;
      | error ->
	  close_in f;
	  raise error
  );
  (* Now [!l] are the packages from this site. Merge into 
   * available_current: 
   *)
  List.iter
    (fun (name, (dir, version, site)) ->
       let newer =
	 try
	   let (_, version', _) = Hashtbl.find available_current name in
	   let (_, v, rev)   = split_name_version ("x-" ^ version) in
	   let (_, v', rev') = split_name_version ("x-" ^ version') in
	   let d =
	     match cmp_version_string v v' with
		 0 -> rev - rev'
	       | n -> n in
	   d > 0
	 with
	     Not_found -> true in
       if newer then
	 Hashtbl.replace available_current name (dir, version, site)
    )
    !l;
  ()
;;


let update_available con =
  let out = printmsg con in

  Hashtbl.clear available_current;
  let makefile = get_tools_makefile() in

  (* Get [sites], the list of GODI sites with build instructions: *)
  let b_sites = Buffer.create 100 in
  exec_to_user_error
    ~pref_user_error:true
    (call_cmd
       ~stdout:(`Buffer b_sites)
       ~stderr:con.protocol
       ~interactivity:con.interactivity
       ~cd:(buildbase()))
    ("godi_console", [| "make"; "-f"; makefile; "get_sites" |]);
  let sites = Str.split words_re (Buffer.contents b_sites) in

  (* For every site, download "available.new", and merge its
   * contents with the variable available_current
   *)
  List.iter
    (fun site ->
       (* Download available.new for [site]: *)
       exec_to_user_error
         ~pref_user_error:true
         (fun () ->
	    (* TODO: pass interactivity *)
	    let cleanup = ref (fun () -> ()) in
	    try
	      let hdl = Godi_file_handler.handler_of_directory site in
	      let (file, clean_opt) = hdl#get con "available.new" in
	      ( match clean_opt with
		  | Some f -> cleanup := f
		  | None -> ()
	      );
	      parse_available con site file;
	      !cleanup()
	    with
	      | File_handler_error msg ->
		  out ("Error: Cannot retrieve package list from " ^ site ^ ": " ^ msg ^ " (ignored)\n");
		  !cleanup()
	 )
         ();
    )
    sites
;;


let install_build_instructions_for con site pkgname version pkgrev =

  let out = printmsg con in

  let cleanup = ref (fun () -> ()) in

  try
    let pkgfullname = 
      Godi_versions.concat_name_version pkgname version pkgrev in
    out ("===> Updating build instructions of " ^ pkgfullname ^ "\n");

    let makefile = get_tools_makefile() in
    let buildfiles = Filename.concat (buildbase()) "buildfiles" in
    
    Godi_file.create_dir ~recurse:true buildfiles;

    (* TODO: pass interactivity *)
    let hdl = Godi_file_handler.handler_of_directory site in
    let tgz = pkgfullname ^ ".build.tgz" in
    let (file, clean_opt) = hdl#get con tgz in
    ( match clean_opt with
	| Some f -> cleanup := f
	| None -> ()
    );
    Godi_file.copy file (Filename.concat buildfiles tgz);

    !cleanup();
    cleanup := (fun () -> ());

    (* TODO: Call TAR directly instead of using console.mk *)

    let args =
      [| "-f"; makefile;
	 "extract_build_instructions";
	 "pkgname=" ^ pkgname;
	 "version=" ^ version;
	 "pkgrev=" ^ string_of_int pkgrev;
      |] in

    exec_to_user_error
      ~pref_user_error:true
      (call_cmd
	 ~stdout:con.protocol
	 ~stderr:con.protocol
	 ~interactivity:con.interactivity)
      ("godi_console", (Array.append [| "make" |] args))
  with
    | File_handler_error msg ->
	!cleanup();
	raise(User_error ("Failed to update build instructions: " ^ msg))
    | error ->
	!cleanup();
	raise error

;;


let update_build_instructions_for con pkgname =
  let (dir,version_pkgrev,site) =
    try
      Hashtbl.find available_current pkgname
    with
	Not_found ->
	  user_error ("No such package: " ^ pkgname) in

  let (_, avail_version, avail_pkgrev) =
    split_name_version ("x-" ^ version_pkgrev) in

  (* Now, if [version] is newer than the installed instance, the
   * build.tgz must be downloaded and extracted.
   *)
  let src_pkgs = scan_source_packages con in
  let is_newer =
    try
      let inst_pkg = List.find (fun pkg ->
				  (source_pkg pkg)#name = pkgname) src_pkgs in
      let inst_version = (source_pkg inst_pkg) # version in
      let inst_pkgrev  = (source_pkg inst_pkg) # pkgrevision in

      let n1 = cmp_version_string avail_version inst_version in
      let n2 = if n1=0 then avail_pkgrev - inst_pkgrev else n1 in

      n2 > 0

    with
	Not_found -> true in

  if is_newer then
    install_build_instructions_for
      con site pkgname avail_version avail_pkgrev
;;


let update_build_instructions con =
  Hashtbl.iter
    (fun pkgname _ ->
       update_build_instructions_for con pkgname)
    available_current
;;
